<?php

namespace App\Models;

use Carbon\Carbon;
use Clarkeash\Doorman\Facades\Doorman;
use Clarkeash\Doorman\Models\Invite;
use Illuminate\Database\Eloquent\Model;

use App\Traits\NextKey;
use App\Traits\PushNew;
use Illuminate\Support\Arr;

class Invoice extends Model
{
	use NextKey,PushNew;
	const RECORD_ID = 'invoice';
	public $incrementing = FALSE;

	protected $table = 'ab_invoice';
	const CREATED_AT = 'date_orig';
	const UPDATED_AT = 'date_last';

	protected $dates = ['date_orig','due_date'];
	public $dateFormat = 'U';

	// Array of items that can be updated with PushNew
	protected $pushable = ['items'];

	protected $with = [
		'account.country.currency',
		'items.taxes',
		'paymentitems'
	];

	private $_total = 0;
	private $_total_tax = 0;

	public function account()
	{
		return $this->belongsTo(Account::class);
	}

	public function items()
	{
		return $this->hasMany(InvoiceItem::class)->where('active',1);
	}

	public function paymentitems()
	{
		return $this->hasMany(PaymentItem::class);
	}

	/** SCOPES **/

	/**
	 * Search for a record
	 *
	 * @param        $query
	 * @param string $term
	 * @return
	 */
	public function scopeSearch($query,string $term)
	{
		return $query->where('id','like','%'.$term.'%');
	}

	/** ATTRIBUTES **/

	public function getDueAttribute()
	{
		return sprintf('%3.'.$this->currency()->rounding.'f',$this->total - $this->paid);
	}

	public function getDateDueAttribute()
	{
		return $this->due_date->format('Y-m-d');
	}

	public function getInvoiceDateAttribute()
	{
		return $this->date_orig->format('Y-m-d');
	}

	public function getInvoiceAccountIdAttribute()
	{
		return sprintf('%02s-%04s-%06s',$this->site_id,$this->account_id,$this->invoice_id);
	}

	public function getInvoiceTextAttribute()
	{
		return sprintf('Thank you for using %s for your Internet Services.',config('SITE_SETUP')->site_name);
	}

	public function getPaidAttribute()
	{
		return $this->currency()->round(
			$this->paymentitems
				->filter(function($item) { return ! $item->payment->pending_status; })
				->sum('alloc_amt'));
	}

	public function getPendingPaidAttribute()
	{
		return $this->currency()->round(
			$this->paymentitems
				->filter(function($item) { return $item->payment->pending_status; })
				->sum('alloc_amt'));
	}

	public function getSubTotalAttribute()
	{
		return sprintf('%3.'.$this->currency()->rounding.'f',$this->total-$this->tax_total);
	}

	public function getTaxTotalAttribute()
	{
		if (! $this->_total_tax)
		{
			foreach ($this->items as $o)
			{
				if ($o->active)
					$this->_total_tax += $this->currency()->round($o->tax);
			}
		}

		return sprintf('%3.'.$this->currency()->rounding.'f',$this->_total_tax);
	}

	public function getTotalAttribute()
	{
		if (! $this->_total)
		{
			foreach ($this->items as $o)
			{
				if ($o->active)
					$this->_total += $this->currency()->round($o->total);
			}
		}

		return sprintf('%3.'.$this->currency()->rounding.'f',$this->_total);
	}

	public function currency()
	{
		return $this->account->country->currency;
	}

	/**
	 * Return a download link for non-auth downloads
	 *
	 * @return string
	 */
	public function download_link(): string
	{
		// Re-use an existing code
		$io = Invite::where('for',$this->account->user->email)->first();

		$tokendate = ($x=Carbon::now()->addDays(21)) > ($y=$this->due_date->addDays(21)) ? $x : $y;

		// Extend the expire date
		if ($io AND ($tokendate > $io->valid_until)) {
			$io->valid_until = $tokendate;
			$io->save();
		}

		$code = (! $io) ? Doorman::generate()->for($this->account->user->email)->uses(0)->expiresOn($tokendate)->make()->first()->code : $io->code;

		return url('u/invoice',[$this->id,'email',$code]);
	}

	public function products()
	{
		$return = collect();

		foreach ($this->items->groupBy('product_id') as $o)
		{
			$po = $o->first()->product;
			$po->count = count($o->pluck('service_id')->unique());

			$return->push($po);
		}

		$lo = $this->account->user->language;
		return $return->sortBy(function ($item) use ($lo) {
			return $item->name($lo);
		});
	}

	public function product_services(Product $po)
	{
		$return = collect();

		$this->items->load(['service']);

		foreach ($this->items->filter(function ($item) use ($po) {
			return $item->product_id == $po->id;
		}) as $o)
		{
			$so = $o->service;
			$return->push($so);
		};

		return $return->unique()->sortBy('name');
	}

	public function product_service_items(Product $po,Service $so)
	{
		return $this->items->filter(function ($item) use ($po,$so) {
			return $item->product_id == $po->id AND $item->service_id == $so->id;
		})->filter()->sortBy('item_type');
	}

	/**
	 * @param string $key
	 * @return string
	 * @todo Ugly hack to update reminders
	 */
	public function reminders(string $key) {
		$r = unserialize($this->reminders);

		if (! Arr::get($r,$key)) {
			$r[$key] = time();
			return serialize($r);
		}
	}

	/**
	 * Automatically set our due_date at save time.
	 *
	 * @param array $options
	 * @return bool
	 */
	public function save(array $options = []) {
		// Automatically set the date_due attribute for new records.
		if (! $this->exists AND ! $this->due_date) {
			$this->due_date = $this->items->min('date_start');

			// @todo This 7 days should be sysetm configurable
			if (($x=Carbon::now()->addDay(7)) > $this->due_date)
				$this->due_date = $x;
		}

		return parent::save($options);
	}
}